/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


// popup init: collect current and clicked folder URI
function InitFolderPanePopup()
{ // get the clicked folder URI
  try
  { // get both folder URIs (may be equal)
    const koFS = goMnenhy.folderstore;
    const koTreeView = GetFolderTree().view;
    const kiCurrent  = koTreeView.selection.currentIndex;
    // 'framed'
    const ksCurrent  = koFS.gsCurrentURI = (kiCurrent < 0) ? "" : koTreeView.getResourceAtIndex(kiCurrent).Value;
    // higlighted
    const ksClicked  = koFS.gsClickedURI = GetSelectedFolderURI();

    // the standard inheritance mode is "inherit from parent"
    // this is assumed whenever we don't have any data for a folder
    const kbInherit = !koFS.rdf.ExistNode(ksClicked);
    const koInherit = document.getElementById("mnenhyFolderPaneInherit");
    koInherit.setAttribute("checked", kbInherit);

    // do not copy from here to here ;-)
    const koFromHere = document.getElementById("mnenhyFolderPaneFromHere");
    koFromHere.setAttribute("disabled", ksCurrent == ksClicked);
  }
  catch(oErr)
  {
    goMnenhy.folderstore.dd(oErr);
  }
}


// |
// |  functions local to this context menu
// |

function InheritFromParent()
{ // change the inheritance mode for the clicked folder (this == window)
  // (when we get here, the checkbox state has already been toggled!)
  const koFS = goMnenhy.folderstore;
  koFS.dd("InheritFromParent()");
  const ksClicked   = koFS.gsClickedURI;
  const ksNoInherit = document.getElementById("mnenhyFolderPaneInherit").getAttribute("checked");
  ToggleCurrentMenuItem(true);
  if (ksNoInherit == "true" && goMnenhy.ConfirmService(koFS.gksConfirmInherit, '', window))
  { // save current folder data
    var sDataFolder = koFS.GetDataFolder(window, koFS.gsCurrentURI, true);
    koFS.SaveFolderData(window, sDataFolder);
    // allow inheritance by deleting the now useless data
    koFS.rdf.RemoveContainer(ksClicked);
    // refresh now to show changes
    sDataFolder = koFS.GetDataFolder(window, koFS.gsCurrentURI, true);
    koFS.LoadFolderData(window, sDataFolder);
  }
  else
  { // cancel inheritance by copying parent data (does own refresh)
    koFS.CopyFromParent(window);
  }
  ToggleCurrentMenuItem(false);
}


function CopyFromParent()
{
  goMnenhy.folderstore.CopyFromParent(window);
}
function CopyFromHere()
{
  goMnenhy.folderstore.CopyFromHere(window);
}


function OpenFolderConfig()
{
  const koFS = goMnenhy.folderstore;
  koFS.dd('OpenFolderConfig()');
  ToggleCurrentMenuItem(true);
  // save current folder data
  var sDataFolder = koFS.GetDataFolder(window, koFS.gsCurrentURI, true);
  koFS.SaveFolderData(window, sDataFolder);
  // do folder config
  window.openDialog("chrome://mnenhy-folderstore/content/mnenhy-folderstore-config.xul",
                    "mnenhyFolderstoreConfig",
                    "chrome,modal,centerscreen,resizable");
  ToggleCurrentMenuItem(false);
  // refresh now to show changes
  sDataFolder = koFS.GetDataFolder(window, koFS.gsCurrentURI, true);
  koFS.LoadFolderData(window, sDataFolder);
}


// if we show a dialog while processing a popup menuitem, we need to temporarily hide the popup
function ToggleCurrentMenuItem(abHide)
{
  document.getElementById("folderPaneContext").collapsed = abHide;
}